/* written by John Doug Reynolds, March 2002 */

#include "rgsenergy.h"
#include <PixelsTable.h>
#include <CalServer.h>
#include <Dal.h>
#include "OalServer.h"
#include "Xmm.h"

void rgsenergy()
{
   string dsname = stringParameter("ccdset");
   bool withoffset = booleanParameter("withoffset");
   bool withdiagoffset = booleanParameter("withdiagoffset");
   bool withgain = booleanParameter("withgain");
   bool withcti = booleanParameter("withcti");

   Xmm::Node nodid = static_cast<Xmm::Node>(calServer.state()->node()->get());
   Xmm::Node cnode = static_cast<Xmm::Node>(0);
   Xmm::Node dnode = static_cast<Xmm::Node>(1);

   if ( withoffset ) {
      string offset = stringParameter("offset");
      calServer.ccf()->registerNewCifEntry( offset );
   }

   struct DataSet *dataset = dataSetServer->open( dsname, Dal::Modify );
   calServer << *dataset;
   unsigned ccdid = calServer.state()->ccd()->get();

   Xmm::Instrument instrument =
       Xmm::toInstrument(dataset->attribute("INSTRUME")->asString());
//   Xmm::Node node;
//   string instrumentname = stringParameter("instrument");
    
   struct EnergyCorrector* calEC = 0; calEC = calServer.getAtom(calEC);
   struct CtiCorrector *calCC = 0;
   if (withcti)
      calCC= calServer.getAtom(calCC);

   {
      using namespace rgsPixelsLib;

      struct Table *ptab = dataset->table("PIXELS");
      struct Table *ftab = dataset->table("EXPOSURE");

      if ( !ptab->hasColumn("PI") )
	 ptab->addColumn( "PI", Dal::Real32, "calibrated energy" );

      unsigned flags = PixelsTable::PIX | PixelsTable::PHA | PixelsTable::CPI;
      const PixelsTable pix(ptab,0,flags);

      const DalVecJ frame( ftab->column("FRAME") );
      const DalVecD  time( ftab->column("TIME") );

      // 

      CalReal32Vector calpiV(  pix.calpi.data(), pix.rows );
      CalInt16Vector   rawxV(   pix.rawX.data(), pix.rows );
      CalInt16Vector energyV( pix.energy.data(), pix.rows );
      CalInt8Vector    nodeV(   pix.node.data(), pix.rows );

      CalInt16Vector  rawyV(  pix.rawY.data(), pix.modeIsHTR()? 0 : pix.rows );
      CalInt8Vector  shapeV( pix.shape.data(), pix.modeIsHTR()? 0 : pix.rows );

      CalReal32Vector  tmpV( 0, pix.rows );
      CalReal64Vector timeV( 0, pix.rows );

      unsigned fItr = 0, pItr = 0;
      for (; fItr < ftab->rows() && pItr < pix.rows; ++fItr )
	 if ( frame[fItr] >= pix.frame[pItr] ) {
	    int sameFrame = pix.frame[pItr];
	    double t = time[ frame[fItr] == sameFrame ? fItr : 0 ];
	    while ( pItr < pix.rows && pix.frame[pItr] == sameFrame )
	       timeV[pItr++] = t;
	 }

      while ( pItr < pix.rows ) timeV[pItr++] = time[0];

      // 

//      DataSet *diag = dataSetServer->open("diag.fit",DataSet::Read);
//      Array* arr0 = diag->array( (ccdid-1) * 2 );
//      ArrayData *data0 = arr0->data();
//      Array* arr1 = diag->array( (ccdid-1) * 2 + 1);
//      ArrayData *data1 = arr1->data();	
			       
      unsigned i, c = 0;
   unsigned j = 0;
   unsigned pflag1 = 0;
   unsigned pflag2 = 0;
   unsigned pflag3 = 0;

     static unsigned short trans[] = { 1, 4, 2, 2, 3, 3, 3, 3 }; 
   static char nodec[] = { 'c', 'd' };

//      int16* a0 = data0->int16Data();
//      int16* a1 = data1->int16Data();

 //     unsigned yExtent = arr0->size()[1];

      if ( pix.modeIsHTR() )
	 calEC->offsetCorrect(timeV,rawxV,nodeV,energyV,calpiV);

// CG pixel dependant offset ...........................

      else {
	 if ( withdiagoffset ) {


        oalServer.instrument(instrument);
        oalServer.ccd(ccdid);
        oalServer.node(cnode);
    	SImage diagnosticMapC = oalServer.odf()->rgsDiagnosticMap();
        oalServer.node(dnode);
    	SImage diagnosticMapD = oalServer.odf()->rgsDiagnosticMap();
        
 	unsigned long xExtent = diagnosticMapC.xExtent();
	cout << "xExtent :" << xExtent << endl;
	unsigned long yExtent = diagnosticMapC.yExtent();
	cout << "yExtent :" << yExtent << endl;
    	
	cout<<"***CCD: "<< ccdid << endl; 
   	
	for ( i = 0; i < 10; ++i ) 
		for ( j = 0; j < 10 ; ++j ) 
		  cout<<"***Cnode: "<< diagnosticMapC[j][i] << " i: " << i << " j: " << j <<endl;
//	       	 cout<<"***Dnode: "<< diagnosticMapD[j][yExtent - i -1]<< " i: " << i << " j: " << j  <<endl;
//	   }
//	}

	for ( i = 0; i < pix.rows; ++i ) {

	      if ( pflag2 < 20 ) {
	       cout<<"P** nodid:"<<nodid<<" eventnode:"<<nodec[nodeV[i]]<<" ---"<<endl;
	       cout<<" ** coord: "<<rawxV[i]<<":"<<rawyV[i]<<endl;
	       cout<<" ** npix: "<<trans[shapeV[i]]<<" energy "<<energyV[i]<<endl;
	       pflag2++;
	 if ( nodeV[i] == 0 ) 
		cout<<"***Subtracting: "<< diagnosticMapC[rawyV[i]-1][rawxV[i]-1] <<endl;
	    else
	       	cout<<"***Subtracting: "<< diagnosticMapD[rawyV[i]-1][yExtent - rawxV[i]] <<endl;
	      }

	     if ( pflag3 < 20 && nodeV[i] != nodid ) {
	       cout<<"*** nodid:"<<nodid<<" eventnode:"<<nodec[nodeV[i]]<<" ---"<<endl;
	       cout<<" ** coord: "<<rawxV[i]<<":"<<rawyV[i]<<endl;
	       cout<<" ** npix: "<<trans[shapeV[i]]<<" energy "<<energyV[i]<<endl;
	       pflag3++;
	     }
	     if ( ( rawyV[i] == 0 || rawxV[i] == 0 ) && pflag1 < 20 ) {
	       cout<<"*** coord: "<<rawxV[i]<<":"<<rawyV[i]<<endl;
	       pflag1++;
	     }

	 if ( nodeV[i] == 0 ) 
	       calpiV[i] = energyV[i] - ( trans[shapeV[i]] * 
	       		diagnosticMapC[rawyV[i]-1][rawxV[i]-1]);
//				  a0[(rawyV[i]-1)*yExtent + rawxV[i]-1] );
	    else
	       calpiV[i] = energyV[i] - ( trans[shapeV[i]] *
	       		diagnosticMapD[rawyV[i]-1][yExtent - rawxV[i]]);
//				  a1[(rawyV[i]-1)*yExtent + (yExtent-rawxV[i]) ] );
	    }

        }
//
	else
	
// Single value CAL offset correction 

	   calEC->offsetCorrect(timeV,rawxV,rawyV,shapeV,nodeV,energyV,calpiV);
     
      }

      CalReal32Vector* v[2] = { &calpiV, &tmpV };


      if ( withgain ) {
	 calEC->gainCorrect(timeV,nodeV,*v[c],*v[1-c]);
	 c = 1 - c;
      }

      if ( withcti ) {
	 if ( pix.modeIsHTR() )
	    calCC->ctiCorrect(timeV,rawxV,pix.htrY0,*v[c],*v[1-c],&nodeV);
	 else {
	    CalInt16Vector sepV( 0, pix.rows );
	    for ( i = 0; i < pix.rows; ++i ) sepV[i] = 1000; // dummy value
	    calCC->ctiCorrect(timeV,rawxV,rawyV,shapeV,sepV,
			      *v[c],*v[1-c],&nodeV);
	 }
	 c = 1 - c;
      }

      if ( c ) for ( i = 0; i < pix.rows; ++i ) pix.calpi[i] = tmpV[i];
   }

   dataSetServer->close( dataset );

   if ( withcti ) calServer.releaseAtom(calCC);
   calServer.releaseAtom(calEC);
}
